@@ -15,10 +15,6 @@
 
 """Functions to manage the common assets for domains."""
 
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
-
 import os
 from dm_control.utils import io as resources
 
